from abc import ABC, abstractmethod
from functools import wraps
from typing import Tuple, List, Iterator, Optional

from .common import ReadOnlyException, Mode, Name


class WriteModeMixin:
    _mode = Mode.READ_ONLY

    @property
    def mode(self):
        return self._mode

    def raise_on_readonly(self):
        if not self.mode.writable:
            raise ReadOnlyException(f"Cannot write to readonly {type(self)}")


def mutation(fn):
    @wraps(fn)
    def wrapped(obj: WriteModeMixin, *args, **kwargs):
        obj.raise_on_readonly()
        return fn(obj, *args, **kwargs)

    return wrapped


class H5ObjectLike(WriteModeMixin, ABC):
    _is_file = False

    def __init__(self, mode: Mode = Mode.default()):
        self._mode = Mode.from_str(mode)

    @property
    @abstractmethod
    def attrs(self) -> "AttributeManagerBase":  # noqa
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    def file(self) -> "FileMixin":  # noqa
        p = self
        for p in self._ancestors():
            pass
        return p

    @property
    @abstractmethod
    def parent(self):
        pass

    @property
    def mode(self):
        return self.parent.mode

    def _ancestors(self) -> Iterator["H5ObjectLike"]:
        """Iterate through ancestors (including this object) until the root"""
        ancestor = self
        while not ancestor._is_file:
            yield ancestor
            ancestor = ancestor.parent
        yield ancestor

    def _absolute_name(self, other: str) -> str:
        """Get the absolute name of the object at the given absolute or relative name"""
        name = Name(self.name)
        return str(name.joinpath(other))

    def _relative_name(self, other: str) -> Optional[str]:
        """Get the relative name of the other object from this one.

        Raises ValueError if they cannot be reached; return None if it is this object.
        """
        name = Name(self.name)
        other = self._absolute_name(other)
        out = str(Name(other).relative_to(name))
        return None if out == "." else out

    def _ancestor_and_relative_name(
        self, other: str
    ) -> Tuple["H5ObjectLike", Optional[str]]:
        """Get the most recent common ancestor with the other name,
        and the relative path from that ancestor to the other name"""
        for ancestor in self._ancestors():
            try:
                return ancestor, ancestor._relative_name(other)
            except ValueError as e:
                if "does not start with" not in str(e):
                    raise

        raise RuntimeError(f"Name '{other}' is not relative to root file")

    def _descend(self, other: str) -> Tuple["H5ObjectLike", List[str], Optional[str]]:
        """Get most recent common ancestor and intermediate names to other.

        :param other:
        :return: most recent common ancestor group, list of intermediate names,
            name of other (or None)
        """
        ancestor, relative_name = self._ancestor_and_relative_name(other)
        if not relative_name:
            return ancestor, [], relative_name

        *mids, final = Name(relative_name).parts
        return ancestor, list(mids), final



from abc import ABC
from copy import deepcopy

import numpy as np

from h5py_like import GroupBase, Name
from .common import check_attrs_rw, ds_kwargs, LoggedClassMixin

random_state = np.random.RandomState(2019)


class GroupLikeTestsMixin(LoggedClassMixin, ABC):
    dataset_kwargs = deepcopy(ds_kwargs)

    def test_attrs(self, obj: GroupBase):
        check_attrs_rw(obj.attrs)

    def test_create_group(self, obj: GroupBase):
        name = "another"
        g2 = obj.create_group(name)
        assert g2.parent == obj
        assert g2.name == str(Name(obj.name) / name)
        assert name in obj

    def test_create_dataset(self, obj: GroupBase):
        dsk = self.dataset_kwargs
        ds = obj.create_dataset(**dsk)
        assert ds.name == str(Name(obj.name) / dsk["name"])
        assert ds.shape == dsk["shape"]
        assert ds.dtype == dsk["dtype"]

    def test_create_dataset_from_data(self, obj: GroupBase):
        dsk = deepcopy(self.dataset_kwargs)
        data = np.ones(dsk.pop("shape"), dtype=dsk.pop("dtype"))
        del dsk["name"]
        ds = obj.create_dataset("test_ds", data=data, **dsk)
        np.testing.assert_array_equal(ds[...], data)


class GroupTestBase(GroupLikeTestsMixin, ABC):
    group_name = "group"

    def group(self, parent, name=None, **kwargs):
        if name is None:
            name = self.group_name

        return parent.create_group(name, **kwargs)

    def test_attrs(self, file_: GroupBase):
        super().test_attrs(self.group(file_))

    def test_create_group(self, file_: GroupBase):
        super().test_create_group(self.group(file_))

    def test_create_dataset(self, file_: GroupBase):
        super().test_create_dataset(self.group(file_))

    def test_create_dataset_from_data(self, file_: GroupBase):
        super().test_create_dataset_from_data(self.group(file_))



from .dataset import DatasetTestBase, ThreadedDatasetTestBase
from .group import GroupTestBase
from .file import FileTestBase, ModeTestBase

__all__ = [
    "DatasetTestBase",
    "GroupTestBase",
    "FileTestBase",
    "ModeTestBase",
    "ThreadedDatasetTestBase",
]



from .common import Mode, Name
from .base import mutation
from .group import GroupBase
from .file import FileMixin
from .dataset import DatasetBase
from .attributes import AttributeManagerBase

from .version import __version__, __version_info__  # noqa

__all__ = [
    "GroupBase",
    "FileMixin",
    "DatasetBase",
    "AttributeManagerBase",
    "mutation",
    "Name",
    "Mode",
]



from abc import ABC, abstractmethod
from pathlib import Path

from .common import Mode, classname


class FileMixin(ABC):
    _is_file = True

    @abstractmethod
    def __init__(self, name, mode=Mode.READ_WRITE_CREATE):
        self.filename: Path = Path(name).resolve()
        self._mode: Mode = Mode.from_str(mode)

    @property
    def parent(self) -> "FileMixin":
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.flush()
        self.close()

    def flush(self):
        pass

    def close(self):
        pass

    @property
    def mode(self):
        return self._mode

    @property
    def name(self):
        return "/"

    def __str__(self):
        return f"<{classname(self)}(filename='{self.filename}', mode='{self.mode}')>"

    def __eq__(self, other):
        try:
            return all(
                (
                    isinstance(other, FileMixin),
                    other._is_file,
                    self.filename == other.filename,
                    self.mode == other.mode,
                )
            )
        except AttributeError:
            return True



from enum import Enum
from pathlib import PurePosixPath


class ReadOnlyException(RuntimeError):
    pass


class StrEnum(str, Enum):
    """Enum subclass which members are also instances of str
    and directly comparable to strings. str type is forced at declaration.
    """

    def __new__(cls, *args):
        for arg in args:
            if not isinstance(arg, str):
                raise TypeError("Not text %s:" % arg)
        return super(StrEnum, cls).__new__(cls, *args)

    def __str__(self):
        return self.value

    def __hash__(self):
        return hash(str(self))


class Mode(StrEnum):
    READ_ONLY = "r"
    READ_WRITE = "r+"
    CREATE_TRUNCATE = "w"
    CREATE = "x"
    READ_WRITE_CREATE = "a"

    @classmethod
    def default(cls):
        return cls.READ_ONLY

    @property
    def writable(self):
        return self != type(self).READ_ONLY

    @classmethod
    def from_str(cls, s):
        if s == "w-":
            s = "x"
        return cls(s)

    def __eq__(self, other):
        if not isinstance(other, type(self)) and other == "w-":
            other = "x"
        return super().__eq__(other)


def classname(obj):
    cls = type(obj)
    module = cls.__module__
    name = cls.__qualname__
    if module is not None and module != "__builtin__":
        name = module + "." + name
    return name


Name = PurePosixPath



from abc import ABC
from collections.abc import MutableMapping

import numpy as np

from .common import Mode
from h5py_like.base import WriteModeMixin, mutation


class AttributeManagerBase(MutableMapping, WriteModeMixin, ABC):
    """
    Allows dictionary-style access to an HDF5-like object's attributes.
    These are created exclusively by the library and are available as
    a Python attribute at <object>.attrs
    Like Group objects, attributes provide a minimal dictionary-
    style interface.  Anything which can be reasonably converted to a
    Numpy array or Numpy scalar can be stored.
    Attributes are automatically created on assignment with the
    syntax <obj>.attrs[name] = value, with the type automatically
    deduced from the value.  Existing attributes are overwritten.
    To modify an existing attribute while preserving its (numpy) shape and type, use the
    method modify().  To specify an attribute of a particular (numpy) type and
    shape, use create().
    """

    def __init__(self, mode: Mode = Mode.default()):
        self._mode = Mode.from_str(mode)

    @mutation
    def create(self, name, data, shape=None, dtype=None):
        """ Set a new attribute, overwriting any existing attribute.
        The type and shape of the attribute are determined from the data.  To
        use a specific (numpy) type or shape, or to preserve the type of an attribute,
        use the methods create() and modify().
        """
        if dtype is None:
            dtype = getattr(data, "dtype", None)
        else:
            dtype = np.dtype(dtype)

        arr = np.asarray(data, dtype=dtype)

        if shape is not None and shape != arr.shape:
            arr = arr.reshape(shape)

        self[name] = arr

    @mutation
    def modify(self, name, value):
        """ Change the value of an attribute while preserving its (numpy) type.
        Differs from __setitem__ in that if the attribute already exists, its
        type is preserved.  This can be very useful for interacting with
        externally generated files.
        If the attribute doesn't exist, it will be automatically created.
        """
        existing = self.get(name)
        if not existing:
            self.create(name, value)
        else:
            self.create(
                name,
                value,
                getattr(existing, "shape", None),
                getattr(existing, "dtype", None),
            )



import logging
import numpy as np

from h5py_like.common import classname

ds_kwargs = {"name": "dataset", "shape": (10, 10, 10), "dtype": np.dtype("uint16")}


def check_attrs_rw(attrs):
    assert "key" not in attrs
    attrs["key"] = "value"
    assert "key" in attrs
    assert attrs["key"] == "value"
    del attrs["key"]
    assert "key" not in attrs


class LoggedClassMixin:
    @property
    def logger(self):
        return logging.getLogger(classname(self))



import itertools
import numbers
from abc import ABC, abstractmethod
from multiprocessing.pool import ThreadPool
from math import ceil
from typing import Callable, Tuple, TypeVar, Union, NamedTuple, List, Iterator
import logging

import numpy as np


logger = logging.getLogger(__name__)

DEFAULT_THREADS = 4


class NullSlicingException(ValueError):
    def __init__(self, msg="", out_shape=None):
        super().__init__(msg)
        self.out_shape = out_shape


class StartLenStride(NamedTuple):
    start: int
    len: int
    stride: int


def slice_to_start_len_stride(slice_: slice, max_len: int) -> StartLenStride:
    """
    Convert a slice object, possibly with None or negative members, into positive
    integers for start, length, and stride.

    Raises NullSlicingException if there would be no results returned from this
    indexing.

    If stride is negative, start is still the lower corner and length is still positive:
    the negative stride must be done after the data read.

    :param slice_:
    :param max_len: maximum length of the dimension
    :return: tuple of positive integer start, length, +ve or -ve stride
    :raises: NullSlicingException
    """
    start, old_stop, stride = slice_.indices(max_len)
    stop = max(old_stop, 0)
    if stop > old_stop:
        start += abs(old_stop)
    shape = stop - start

    if shape == 0 or shape * stride < 0:
        raise NullSlicingException("Requested slice is empty")

    return StartLenStride(min(start, stop), abs(shape), stride)


def int_to_start_len_stride(i: int, max_len: int) -> StartLenStride:
    """
    Convert an integer index, possibly negative, into positive integers for start,
    length, and stride.

    Raises NullSlicingException if there would be no results returned from this indexing

    :param i: integer index
    :param max_len: maximum length of the dimension
    :return: tuple of positive integer start, length, stride
    """
    if -max_len < i < 0:
        begin = i + max_len
    elif i >= max_len or i < -max_len:
        raise IndexError(
            "index {} is out of bounds for axis with size".format(i, max_len)
        )
    else:
        begin = i

    return StartLenStride(begin, 1, 1)


class Roi(NamedTuple):
    """ROI described by positive integers.
    """

    start: Tuple[int, ...]
    read_shape: Tuple[int, ...]
    stride: Tuple[int, ...]
    out_shape: Tuple[int, ...]


ConcreteSliceLike = Union[slice, int]
SliceLike = Union[ConcreteSliceLike, type(Ellipsis), np.newaxis]
SliceArgs = TypeVar("SliceArgs", SliceLike, Tuple[SliceLike])


class Indexer:
    type_msg = (
        "Advanced selection inappropriate. "
        "Only numbers, slices (`:`), ellipsis (`...`), and np.newaxis (None) "
        "are valid indices (or tuples thereof); got {}"
    )
    non_indexes = (None, Ellipsis)

    def __init__(self, max_shape: Tuple[int, ...]):
        """Class for normalising some of numpy's indexing scheme.

        :param max_shape: shape of the array to be read
        """
        self.max_shape = max_shape
        self.ndim = len(max_shape)

    def _handle_newaxis_ellipses(
        self, index_tup: Tuple[SliceLike]
    ) -> Tuple[List[ConcreteSliceLike], List[int]]:
        """

        :param index_tup:
        :return: list of ints and slices the same length as the dataset's shape,
         and list of indices at which to insert newaxes (must be in inserted in order)
        """
        concrete_indices = sum(idx not in self.non_indexes for idx in index_tup)
        index_lst = []
        newaxis_at = []
        has_ellipsis = False
        int_count = 0
        for item in index_tup:
            if isinstance(item, numbers.Number):
                int_count += 1

            if item is None:
                newaxis_at.append(len(index_lst) + len(newaxis_at) - int_count)
            elif item == Ellipsis:
                if has_ellipsis:
                    raise IndexError("an index can only have a single ellipsis ('...')")
                has_ellipsis = True
                initial_len = len(index_lst)
                while len(index_lst) + (concrete_indices - initial_len) < self.ndim:
                    index_lst.append(slice(None))
            else:
                index_lst.append(item)

        if len(index_lst) > self.ndim:
            raise IndexError("too many indices for array")
        while len(index_lst) < self.ndim:
            index_lst.append(slice(None))

        return index_lst, newaxis_at

    def __getitem__(self, args: SliceArgs):
        index_tup = np.index_exp[args]
        index_lst, newaxis_at = self._handle_newaxis_ellipses(index_tup)

        start_readshape_stride = []
        out_shape = []

        for item in index_lst:
            d = len(start_readshape_stride)
            max_len = self.max_shape[d]

            if isinstance(item, slice):

                try:
                    bls = slice_to_start_len_stride(item, max_len)
                    start_readshape_stride.append(bls)
                    out_shape.append(ceil(bls.len / abs(bls.stride)))
                except NullSlicingException:
                    start_readshape_stride.append((None, None, None))
                    out_shape.append(0)

            elif isinstance(item, (int, np.integer)):

                start_readshape_stride.append(
                    int_to_start_len_stride(int(item), max_len)
                )

            else:
                raise TypeError(self.type_msg.format(type(item)))

        for newax_idx in newaxis_at:
            out_shape.insert(newax_idx, 1)

        out_shape = tuple(out_shape)
        if 0 in out_shape:
            raise NullSlicingException("Slicing has a 0-length dimension", out_shape)

        start, read_shape, stride = zip(*start_readshape_stride)
        return Roi(start, read_shape, stride, out_shape)


class IndexableArrayLike(ABC):
    @property
    @abstractmethod
    def shape(self) -> Tuple[int]:
        pass

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def size(self) -> int:
        return np.prod(self.shape, dtype=np.intp).item()

    @property
    @abstractmethod
    def dtype(self):
        pass

    @property
    def _indexer(self):
        return Indexer(self.shape)

    def _getitem(
        self,
        args: SliceArgs,
        read_fn: Callable[[Tuple[int, ...], Tuple[int, ...]], np.ndarray],
        dtype: np.dtype = None,
    ) -> np.ndarray:
        """
        Use a given function to get data from an underlying dataset, and stride it with
        numpy if necessary.

        The given function could handle threading; ``thread_read_fn`` is a helper for
        this.

        :param args: arguments as passed to __getitem__
        :param dtype: data type of the data (default whatever is returned by the
            internal function)
        :param read_fn: callable which takes offset and shape tuples of positive
            integers, and returns a numpy array
        :return: numpy array of the returned data with the requested dtype
        """
        dtype = self.dtype if dtype is None else np.dtype(dtype)
        try:
            start, read_shape, stride, out_shape = self._indexer[args]
        except NullSlicingException as e:
            return np.empty(e.out_shape, dtype=dtype)

        arr = np.asarray(read_fn(start, read_shape), dtype=dtype)

        if set(stride) != {1}:
            stride_slices = tuple(slice(None, None, s) for s in stride)
            arr = arr[stride_slices]

        if not out_shape:
            return arr.flatten()[0]
        elif arr.shape != out_shape:
            return arr.reshape(out_shape)
        else:
            return arr

    def _setitem(
        self,
        args: SliceArgs,
        array: np.ndarray,
        write_fn: Callable[[Tuple[int, ...], np.ndarray], None],
    ):
        """
        Use a given function to insert data into an underlying dataset.
        Does not support striding, or broadcasting other than scalar.

        The given function could handle threading; ``thread_write_fn`` is a helper for
        this.

        :param args: index arguments as passed to __setitem__
        :param array: array-like data to write
        :param write_fn: function which takes an offset as a tuple of integers, and a
            numpy array, and writes to an underlying dataset
        """
        try:
            start, host_shape, stride, _ = self._indexer[args]
        except NullSlicingException:
            return

        if set(stride) != {1}:
            raise NotImplementedError("Strided writes are not supported")

        if np.isscalar(array):
            item_arr = np.full(host_shape, array, dtype=self.dtype)
        else:
            try:
                item_arr = np.asarray(array, self.dtype, order="C")
            except ValueError as e:
                if any(
                    s in str(e) for s in ("invalid literal for ", "could not convert")
                ):
                    bad_dtype = np.asarray(array).dtype
                    raise TypeError("No conversion path for dtype: " + repr(bad_dtype))
                else:
                    raise
            except TypeError as e:
                if "argument must be" in str(e):
                    raise OSError(
                        "Can't prepare for writing data "
                        "(no appropriate function for conversion path)"
                    )
                else:
                    raise

        if host_shape != item_arr.shape:
            raise ValueError("Shape of array does not match shape of index")

        return write_fn(start, item_arr)


CHUNK_BASE = 256 * 1024  # Multiplier by which chunks are adjusted
CHUNK_MIN = 128 * 1024  # Soft lower limit (128k)
CHUNK_MAX = 64 * 1024 * 1024  # Hard upper limit


def guess_chunks(shape, typesize):
    """
    Guess an appropriate chunk layout for a dataset, given its shape and
    the size of each element in bytes.  Will allocate chunks only as large
    as MAX_SIZE.  Chunks are generally close to some power-of-2 fraction of
    each axis, slightly favoring bigger values for the last index.
    Undocumented and subject to change without warning.

    Originally implemented in zarr_ (MIT license)

    .. _zarr: https://github.com/zarr-developers/zarr-python/blob/e61d6ae77f18e881be0b80e38b5366793f5a2860/zarr/util.py#L64-L112
    """  # noqa

    ndims = len(shape)
    # require chunks to have non-zero length for all dimensions
    chunks = np.maximum(np.array(shape, dtype="=f8"), 1)

    # Determine the optimal chunk size in bytes using a PyTables expression.
    # This is kept as a float.
    dset_size = np.product(chunks) * typesize
    target_size = CHUNK_BASE * (2 ** np.log10(dset_size / (1024.0 * 1024)))

    if target_size > CHUNK_MAX:
        target_size = CHUNK_MAX
    elif target_size < CHUNK_MIN:
        target_size = CHUNK_MIN

    idx = 0
    while True:
        # Repeatedly loop over the axes, dividing them by 2.  Stop when:
        # 1a. We're smaller than the target chunk size, OR
        # 1b. We're within 50% of the target chunk size, AND
        # 2. The chunk is smaller than the maximum chunk size

        chunk_bytes = np.product(chunks) * typesize

        if (
            chunk_bytes < target_size
            or abs(chunk_bytes - target_size) / target_size < 0.5
        ) and chunk_bytes < CHUNK_MAX:
            break

        if np.product(chunks) == 1:
            break  # Element size larger than CHUNK_MAX

        chunks[idx % ndims] = np.ceil(chunks[idx % ndims] / 2.0)
        idx += 1

    return tuple(int(x) for x in chunks)


def to_slice(start: Tuple[int, ...], shape: Tuple[int, ...]):
    """Convert start and shape tuple to slice object"""
    return tuple(slice(st, st + sh) for st, sh in zip(start, shape))


def chunk_roi(
    start: Tuple[int, ...],
    shape: Tuple[int, ...],
    chunks: Tuple[int, ...],
    global_shape: Tuple[int, ...],
) -> Iterator[Tuple[Tuple[int, ...], Tuple[int, ...]]]:
    """Decompose a large ROI into chunks.

    Yields a (start, shape) pair of tuples for each subregion.
    Given the chunking scheme ``chunks``, each subregion is constrained to a single
    block, and there is only a single subregion per block.
    Subregions may be smaller than a block if the input ``start``, ``shape`` pair is not
    block-aligned.

    :param start: of the ROI
    :param shape: of the ROI
    :param chunks: chunking scheme of the array which will be indexed into
    :param global_shape: shape of the array which will be indexed into
    :return: start, shape pairs for each subregion
    """
    chunks = np.asarray(chunks, dtype=int)

    start = np.asarray(start, dtype=int)
    shape = np.asarray(shape, dtype=int)
    stop = start + shape

    start_block_idx, start_block_offset = divmod(start, chunks)

    last_block_idx, last_block_offset = divmod(stop, chunks)
    last_block_idx[last_block_offset == 0] -= 1
    last_block_offset[last_block_offset == 0] = chunks[last_block_offset == 0]

    internal_block_idxs = last_block_idx - start_block_idx

    for internal_block_idx in itertools.product(
        *(range(d) for d in internal_block_idxs + 1)
    ):
        internal_block_idx = np.asarray(internal_block_idx, dtype=int)
        global_block_idx = start_block_idx + internal_block_idx

        s_b_offset = start_block_offset * (internal_block_idx == 0)
        l_b_offset = chunks.copy()

        l_b_offset[global_block_idx == last_block_idx] = last_block_offset[
            global_block_idx == last_block_idx
        ]

        global_block_start_coords = global_block_idx * chunks
        start_coords = global_block_start_coords + s_b_offset

        this_shape = tuple(
            min(gbsc + lbo, gs) - sc
            for gbsc, lbo, gs, sc in zip(
                global_block_start_coords, l_b_offset, global_shape, start_coords
            )
        )

        yield tuple(start_coords), this_shape


def thread_read_fn(
    start: Tuple[int, ...],
    shape: Tuple[int, ...],
    chunks: Tuple[int, ...],
    global_shape: Tuple[int, ...],
    read_fn: Callable[[Tuple[int, ...], Tuple[int, ...]], np.ndarray],
    threads: int = DEFAULT_THREADS,
):
    """For chunked datasets, split reading across threads.

    :param start:
    :param shape:
    :param chunks: chunking scheme of array to be read
    :param global_shape: shape of the array to be read
    :param read_fn: function which takes a ``start`` and ``shape`` tuple,
        and returns an array
    :param threads:
    :return:
    """
    source_coords_list = list(chunk_roi(start, shape, chunks, global_shape))
    threads = min(threads, len(source_coords_list))
    with ThreadPool(threads) as pool:
        logger.debug("Started pool of %s threads for reading", threads)
        results = pool.starmap(
            read_fn, source_coords_list, chunksize=len(source_coords_list) // threads
        )

    arr = np.empty(shape, dtype=results[0].dtype)

    for result, (start_coords, block_shape) in zip(results, source_coords_list):
        slicing = to_slice(np.array(start_coords) - start, block_shape)
        arr[slicing] = result

    return arr


def thread_write_fn(
    start: Tuple[int, ...],
    arr: np.ndarray,
    chunks: Tuple[int, ...],
    global_shape: Tuple[int, ...],
    write_fn: Callable[[Tuple[int, ...], np.ndarray], None],
    threads=DEFAULT_THREADS,
):
    """

    :param start:
    :param arr: array to write to the ROI
    :param chunks: chunking scheme of array to be written to
    :param global_shape: shape of array to be written to
    :param write_fn: function which takes a ``start`` tuple and a np.ndarray
    :param threads:
    :return:
    """
    tgt_coords_list = list(chunk_roi(start, arr.shape, chunks, global_shape))

    threads = min(threads, len(tgt_coords_list))
    with ThreadPool(min(threads, len(tgt_coords_list))) as pool:
        logger.debug("Started pool of %s threads for writing", threads)
        pool.starmap(
            write_fn,
            (
                [st, arr[to_slice(np.array(st) - start, sh)]]
                for st, sh in tgt_coords_list
            ),
            chunksize=len(tgt_coords_list) // threads,
        )



from abc import ABC, abstractmethod
from collections.abc import MutableMapping

import numpy as np
from typing import Any, Optional, Callable

from .base import H5ObjectLike, mutation
from .dataset import DatasetBase
from .common import classname


class GroupBase(H5ObjectLike, MutableMapping, ABC):
    """ Represents an HDF5-like group.
    """

    _is_file = False

    @abstractmethod
    def _create_child_group(self, name) -> "GroupBase":
        """Create a group which is a direct child of this object with the given name.

        Should raise a TypeError if a dataset exists there,
        or a FileExistsError if a group exists there.
        """
        pass

    def _require_descendant_groups(self, *names):
        this = self
        for name in names:
            try:
                this = this._get_child(name)
                if not isinstance(this, GroupBase):
                    raise TypeError("Not a group")
            except KeyError:
                this = this._create_child_group(name)
        return this

    @mutation
    def create_group(self, name) -> "GroupBase":
        """ Create and return a new subgroup.
        Name may be absolute or relative.  Fails if the target name already
        exists.
        """
        ancestor, group_names, last_name = self._descend(name)
        parent = ancestor._require_descendant_groups(*group_names)
        if last_name in parent:
            raise FileExistsError(f"Group or dataset found at '{name}'")
        return parent._create_child_group(last_name)

    @abstractmethod
    def _create_child_dataset(self, name, shape=None, dtype=None, data=None, **kwds):
        """Create a dataset which is a direct child of this object with the given name.

        Should raise a TypeError if a group exists there,
        or a FileExistsError if a dataset exists there.
        """
        pass

    @mutation
    def create_dataset(
        self, name, shape=None, dtype=None, data=None, **kwds
    ) -> DatasetBase:
        """ Create a new HDF5-like dataset

        name
            Name of the dataset (absolute or relative).
        shape
            Dataset shape.  Use "()" for scalar datasets.  Required if "data"
            isn't provided.
        dtype
            Numpy dtype or string.  If omitted, dtype('f') will be used.
            Required if "data" isn't provided; otherwise, overrides data
            array's dtype.
        data
            Provide data to initialize the dataset.  If used, you can omit
            shape and dtype arguments.

        If the dataset implementation supports chunking:

        chunks
            tuple describing the shape of the chunks
        """
        ancestor, group_names, last_name = self._descend(name)
        parent = ancestor._require_descendant_groups(*group_names)
        if last_name in parent:
            raise FileExistsError(f"Group or dataset found at '{name}'")
        return parent._create_child_dataset(name, shape, dtype, data, **kwds)

    def require_dataset(self, name, shape, dtype, exact=False, **kwds) -> DatasetBase:
        """ Open a dataset, creating it if it doesn't exist.
        If keyword "exact" is False (default), an existing dataset must have
        the same shape and a conversion-compatible dtype to be returned.  If
        True, the shape and dtype must match exactly.
        Other dataset keywords (see create_dataset) may be provided, but are
        only used if a new dataset is to be created.
        Raises TypeError if an incompatible object already exists, or if the
        shape or dtype don't match according to the above rules.
        """
        if name not in self:
            return self.create_dataset(name, *(shape, dtype), **kwds)

        dset = self[name]
        if not isinstance(dset, DatasetBase):
            raise TypeError(
                "Incompatible object (%s) already exists" % dset.__class__.__name__
            )

        if not shape == dset.shape:
            raise TypeError(
                "Shapes do not match (existing %s vs new %s)" % (dset.shape, shape)
            )

        if exact:
            if not dtype == dset.dtype:
                raise TypeError(
                    "Datatypes do not exactly match (existing %s vs new %s)"
                    % (dset.dtype, dtype)
                )
        elif not np.can_cast(dtype, dset.dtype):
            raise TypeError(
                "Datatypes cannot be safely cast (existing %s vs new %s)"
                % (dset.dtype, dtype)
            )

        return dset

    def create_dataset_like(self, name, other, **kwupdate) -> DatasetBase:
        """ Create a dataset similar to `other`.
        name
            Name of the dataset (absolute or relative).  Provide None to make
            an anonymous dataset.
        other
            The dataset which the new dataset should mimic. All properties, such
            as shape, dtype, chunking, ... will be taken from it, but no data
            or attributes are being copied.
        Any dataset keywords (see create_dataset) may be provided, including
        shape and dtype, in which case the provided values take precedence over
        those from `other`.
        """
        for k in ("shape", "dtype", "chunks", "fillvalue"):
            kwupdate.setdefault(k, getattr(other, k))

        # Special case: the maxshape property always exists, but if we pass it
        # to create_dataset, the new dataset will automatically get chunked
        # layout. So we copy it only if it is different from shape.
        if other.maxshape != other.shape:
            kwupdate.setdefault("maxshape", other.maxshape)

        return self.create_dataset(name, **kwupdate)

    def require_group(self, name) -> "GroupBase":
        """Return a group, creating it if it doesn't exist.
        TypeError is raised if something with that name already exists that
        isn't a group.
        """
        if name not in self:
            return self.create_group(name)

        group = self[name]
        if not isinstance(group, GroupBase):
            raise TypeError(f"Incompatible object ({type(group)}) already exists")

        return group

    @abstractmethod
    def _get_child(self, name) -> H5ObjectLike:
        """Get an object (a concrete Dataset or Group) which is a direct child of this object.

        Raise a KeyError if it's not found.
        """
        pass

    def __getitem__(self, name) -> H5ObjectLike:
        """ Open an object in the file """
        ancestor, group_names, last_name = self._descend(name)
        for group_name in group_names:
            ancestor = ancestor._get_child(group_name)
            if not isinstance(ancestor, GroupBase):
                raise TypeError(f"Expected Group, got {type(ancestor)}")
        return ancestor._get_child(last_name)

    @abstractmethod
    def __setitem__(self, name, obj):
        """ Add an object to the group.  The name must not already be in use.
        The action taken depends on the type of object assigned:
        Named HDF5 object (Dataset, Group, Datatype)
            A hard link is created at "name" which points to the
            given object.
        SoftLink or ExternalLink
            Create the corresponding link.
        Numpy ndarray
            The array is converted to a dataset object, with default
            settings (contiguous storage, etc.).
        Numpy dtype
            Commit a copy of the datatype as a named datatype in the file.
        Anything else
            Attempt to convert it to an ndarray and store it.  Scalar
            values are stored as scalar datasets. Raise ValueError if we
            can't understand the resulting array dtype.
        """

    @abstractmethod
    def copy(
        self,
        source,
        dest,
        name=None,
        shallow=False,
        expand_soft=False,
        expand_external=False,
        expand_refs=False,
        without_attrs=False,
    ):
        """Copy an object or group.
        The source can be a path, Group, Dataset, or Datatype object.  The
        destination can be either a path or a Group object.  The source and
        destinations need not be in the same file.
        If the source is a Group object, all objects contained in that group
        will be copied recursively.
        When the destination is a Group object, by default the target will
        be created in that group with its current name (basename of obj.name).
        You can override that by setting "name" to a string.
        There are various options which all default to "False":
         - shallow: copy only immediate members of a group.
         - expand_soft: expand soft links into new objects.
         - expand_external: expand external links into new objects.
         - expand_refs: copy objects that are pointed to by references.
         - without_attrs: copy object without copying attributes.
       Example:
        >>> f = File('myfile.hdf5')
        >>> f.listnames()
        ['MyGroup']
        >>> f.copy('MyGroup', 'MyCopy')
        >>> f.listnames()
        ['MyGroup', 'MyCopy']
        """

    @mutation
    def move(self, source, dest):
        """ Move a link to a new location in the file.
        If "source" is a hard link, this effectively renames the object.  If
        "source" is a soft or external link, the link itself is moved, with its
        value unmodified.
        """
        self.copy(source, dest)
        del self[source]

    def _recurse(self):
        """Depth-first search"""
        for k, v in self.items():
            yield k, v
            try:
                yield from v._recurse()
            except AttributeError:
                pass

    def visit(self, func: Callable[[str], Optional[Any]]) -> Optional[Any]:
        """ Recursively visit all names in this group and subgroups (HDF5 1.8).
        You supply a callable (function, method or callable object); it
        will be called exactly once for each link in this group and every
        group below it. Your callable must conform to the signature:
            func(<member name>) => <None or return value>
        Returning None continues iteration, returning anything else stops
        and immediately returns that value from the visit method.  No
        particular order of iteration within groups is guaranteed.
        Example:
        >>> # List the entire contents of the file
        >>> f = File("foo.hdf5")
        >>> list_of_names = []
        >>> f.visit(list_of_names.append)
        """
        for key, _ in self._recurse():
            result = func(key)
            if result is not None:
                return result

    def visititems(
        self, func: Callable[[str, H5ObjectLike], Optional[Any]]
    ) -> Optional[Any]:
        """ Recursively visit names and objects in this group (HDF5 1.8).
        You supply a callable (function, method or callable object); it
        will be called exactly once for each link in this group and every
        group below it. Your callable must conform to the signature:
            func(<member name>, <object>) => <None or return value>
        Returning None continues iteration, returning anything else stops
        and immediately returns that value from the visit method.  No
        particular order of iteration within groups is guaranteed.
        Example:
        # Get a list of all datasets in the file
        >>> mylist = []
        >>> def func(name, obj):
        ...     if isinstance(obj, DatasetBase):
        ...         mylist.append(name)
        ...
        >>> f = File('foo.hdf5')
        >>> f.visititems(func)
        """
        for key, val in self._recurse():
            result = func(key, val)
            if result is not None:
                return result

    def __str__(self):
        return f"<{classname(self)}(name='{self.name}', file={self.file})>"

    def __eq__(self, other):
        try:
            return all(
                (
                    isinstance(other, GroupBase),
                    not other._is_file,
                    self.name == other.name,
                    self.parent == other.parent,
                )
            )
        except AttributeError:
            return False



from setuptools import find_packages, setup


with open("README.md") as f:
    readme = f.read()

extras = {"test": ["pytest>=4.6.3"]}

setup(
    author="Chris Lloyd Barnes",
    author_email="barnesc@janelia.hhmi.org",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
    ],
    description="Abstract base classes for making h5py-like objects.",
    install_requires=["numpy"],
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    extras_require=extras,
    keywords="h5py",
    name="h5py_like",
    packages=find_packages(exclude=["tests"]),
    url="https://github.com/barnesc/h5py_like",
    version="0.5.2",
    zip_safe=False,
)



from abc import ABC, abstractmethod

import pytest

from h5py_like import Mode, FileMixin
from h5py_like.common import ReadOnlyException
from h5py_like.test_utils.common import LoggedClassMixin
from .group import GroupLikeTestsMixin


class FileTestBase(GroupLikeTestsMixin, ABC):
    def cls_ext(self, file_obj):
        return type(file_obj), file_obj.filename.suffix

    def test_attrs(self, file_):
        super().test_attrs(file_)

    def test_create_group(self, file_):
        super().test_create_group(file_)

    def test_create_dataset(self, file_):
        super().test_create_dataset(file_)

    def test_create_dataset_from_data(self, file_):
        super().test_create_dataset_from_data(file_)


class ModeTestBase(LoggedClassMixin, ABC):
    @abstractmethod
    def factory(self, mode: Mode) -> FileMixin:
        """Create a File at a location deterministic per test.
        """
        pass

    def ensure_exists(self, group=None):
        try:
            with self.factory(Mode.READ_WRITE_CREATE) as f:
                if group is not None:
                    f.create_group(group)
        except FileExistsError:
            pass

    def test_read_write_create(self):
        mode = Mode.READ_WRITE_CREATE

        with self.factory(mode) as f:
            f.create_group("group")

        with self.factory(mode) as f2:
            assert "group" in f2

    def test_create(self):
        mode = Mode.CREATE

        self.factory(mode)

        with pytest.raises(FileExistsError):
            with self.factory(mode):
                pass

    def test_read_only(self):
        mode = Mode.READ_ONLY
        with pytest.raises(FileNotFoundError):
            self.factory(mode)

        self.ensure_exists()

        with self.factory(mode) as f:
            with pytest.raises(ReadOnlyException):
                f.create_group("group")

    def test_create_truncate(self):
        mode = Mode.CREATE_TRUNCATE

        self.factory(mode)

        self.ensure_exists("group")

        with self.factory(mode) as f:
            assert "group" not in f

    def test_read_write(self):
        mode = Mode.READ_WRITE

        with pytest.raises(FileNotFoundError):
            self.factory(mode)

        self.ensure_exists()

        with self.factory(mode) as f:
            f.create_group("group")



from abc import ABC
from copy import deepcopy

import numpy as np
import pytest

from h5py_like.shape_utils import to_slice
from .common import check_attrs_rw, ds_kwargs, LoggedClassMixin

random_state = np.random.RandomState(1991)


class DatasetTestBase(LoggedClassMixin, ABC):
    dataset_kwargs = deepcopy(ds_kwargs)

    def dataset(self, parent, data=None, **kwargs):
        kwds = deepcopy(self.dataset_kwargs)
        kwds.update(kwargs)

        if data is not None:
            for key in ("shape", "dtype"):
                del kwds[key]
        kwds["data"] = data

        return parent.create_dataset(**kwds)

    def test_attrs(self, file_):
        check_attrs_rw(self.dataset(file_).attrs)

    def test_simple_rw(self, file_):
        ds = self.dataset(file_)
        data = np.ones_like(ds)
        ds[...] = data
        np.testing.assert_equal(ds[...], data)

    @pytest.mark.parametrize(
        "slice_args",
        [
            np.index_exp[...],
            np.index_exp[::-1, ::-1, ::-1],
            np.index_exp[4:8, 1:19, -5:-1],
            np.index_exp[2, 2, 2],
            np.index_exp[:, :, np.newaxis, :],
            np.index_exp[1:5],
            np.index_exp[..., 1:5],
            np.index_exp[..., 1:5, ...],
        ],
    )
    def test_slicing(self, slice_args, file_):
        data = random_state.random_sample((20, 20, 20))
        ds = self.dataset(file_, data=data)

        impl_e = None
        np_e = None

        try:
            impl = ds[slice_args]
        except Exception as e:
            impl_e = e

        try:
            numpy = data[slice_args]
        except Exception as e:
            np_e = e

        if np_e or impl_e:
            assert isinstance(impl_e, type(np_e))
        else:
            np.testing.assert_allclose(impl, numpy)

    def test_broadcast_scalar(self, file_):
        ds = self.dataset(file_)
        ds[:] = 5
        expected = np.full_like(ds, 5)
        np.testing.assert_allclose(ds, expected)

    @pytest.mark.parametrize(
        "slice_args,total",
        [(np.index_exp[...], 10 ** 3), (np.index_exp[:5, :5, :5], 5 ** 3)],
    )
    def test_setitem_total(self, slice_args, total, file_):
        ds = self.dataset(file_)
        ds[slice_args] = 1
        assert ds[:].sum() == total

    def test_setitem(self, file_):
        data = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])

        ds = self.dataset(file_, data=data, chunks=(3, 1))
        ds[:2, :2] = np.array([[5, 6], [7, 8]], dtype=data.dtype)

        np.testing.assert_allclose(
            ds[:], np.array([[5, 6, 3], [7, 8, 3], [1, 2, 3]], dtype=data.dtype)
        )


chunked = deepcopy(ds_kwargs)
chunked["chunks"] = (5, 5, 5)


class ThreadedDatasetTestBase(DatasetTestBase):
    dataset_kwargs = deepcopy(chunked)

    @pytest.mark.parametrize("threads", [1, 4], ids=lambda x: f"threads{x}")
    def test_threaded_read(self, file_, threads):
        shape = (20, 20)
        chunks = (10, 10)
        data = np.arange(np.product(shape), dtype=int).reshape(shape)
        ds = self.dataset(file_, data, chunks=chunks)
        ds.threads = threads

        read_start = (5, 5)
        read_shape = (10, 10)

        slicing = to_slice(read_start, read_shape)
        assert np.array_equal(ds[slicing], data[slicing])

    @pytest.mark.parametrize("threads", [1, 4], ids=lambda x: f"threads{x}")
    def test_threaded_write(self, file_, threads):
        shape = (20, 20)
        chunks = (10, 10)

        data = np.ones(shape, dtype=int)

        ds = self.dataset(file_, data, chunks=chunks)
        ds.threads = threads

        write_start = (5, 5)
        write_shape = (10, 10)
        write_slice = to_slice(write_start, write_shape)

        data[write_slice] = 9
        ds[write_slice] = 9

        assert np.array_equal(data, ds[:])



import sys
from abc import abstractmethod, ABC
from contextlib import contextmanager

import numpy as np
from typing import Tuple, Optional, Any, Union, Iterator

from .shape_utils import IndexableArrayLike
from .common import Mode, classname
from h5py_like.base import H5ObjectLike, mutation


class DatasetBase(H5ObjectLike, IndexableArrayLike, ABC):
    """Represents an HDF5-like dataset

    If the dataset implementation supports threaded reads and writes, the number of
    threads should be controlled by the ``threads`` attribute.
    """

    threads = None
    _is_file = False

    def __init__(self, mode: Mode = Mode.default()):
        self._astype = None
        super().__init__(mode)

    @contextmanager
    def astype(self, dtype):
        """ Get a context manager allowing you to perform reads to a
        different destination type, e.g.:
        >>> with dataset.astype('f8'):
        ...     double_precision = dataset[0:100:2]
        """
        self._astype = np.dtype(dtype)
        yield
        self._astype = None

    @property
    @abstractmethod
    def dims(self):
        """ Access dimension scales attached to this dataset. """
        raise NotImplementedError()

    @property
    def ndim(self) -> int:
        """Numpy-style attribute giving the number of dimensions"""
        return len(self.shape)

    @property
    @abstractmethod
    def shape(self) -> Tuple[int, ...]:
        """Numpy-style shape tuple giving dataset dimensions"""
        raise NotImplementedError()

    @shape.setter
    def shape(self, shape) -> None:
        self.resize(shape)

    @property
    def size(self) -> int:
        """Numpy-style attribute giving the total dataset size"""
        return np.prod(self.shape, dtype=np.intp).item()

    @property
    @abstractmethod
    def dtype(self) -> np.dtype:
        """Numpy dtype representing the datatype"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def maxshape(self) -> Tuple[int, ...]:
        """Shape up to which this dataset can be resized.  Axes with value
        None have no resize limit. """
        raise NotImplementedError()

    @property
    @abstractmethod
    def fillvalue(self) -> Any:
        """Fill value for this dataset (0 by default)"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def chunks(self) -> Optional[Tuple[int, ...]]:
        """Dataset chunks (or None)"""
        raise NotImplementedError()

    def _sanitize_resize(
        self, size: Union[int, Tuple[int, ...]], axis: Optional[int] = None
    ) -> Tuple[int, ...]:
        if self.chunks is None:
            raise TypeError("Only chunked datasets can be resized")

        if axis is not None:
            if not (0 <= axis < self.ndim):
                raise ValueError("Invalid axis (0 to %s allowed)" % (self.ndim - 1))
            try:
                newlen = int(size)
            except TypeError:
                raise TypeError("Argument must be a single int if axis is specified")
            size = list(self.shape)
            size[axis] = newlen

        return tuple(size)

    @abstractmethod
    def resize(self, size: Union[int, Tuple[int, ...]], axis: Optional[int] = None):
        """ Resize the dataset, or the specified axis.
        The dataset must be stored in chunked format; it can be resized up to
        the "maximum shape" (keyword maxshape) specified at creation time.
        The rank of the dataset cannot be changed.
        "Size" should be a shape tuple, or if an axis is specified, an integer.
        BEWARE: This functions differently than the NumPy resize() method!
        The data is not "reshuffled" to fit in the new shape; each axis is
        grown or shrunk independently.  The coordinates of existing data are
        fixed.

        self._sanitize_resize may help validate the resize operation.
        """
        # checked_size = self._sanitize_resize(size, axis)
        raise NotImplementedError()

    def __len__(self) -> int:
        """ The size of the first axis.  TypeError if scalar.
        Limited to 2**32 on 32-bit systems; Dataset.len() is preferred.
        """
        size = self.len()
        if size > sys.maxsize:
            raise OverflowError(
                "Value too big for Python's __len__; use Dataset.len() instead."
            )
        return size

    def len(self) -> int:
        """ The size of the first axis.  TypeError if scalar.
        Use of this method is preferred to len(dset), as Python's built-in
        len() cannot handle values greater then 2**32 on 32-bit systems.
        """
        shape = self.shape
        if len(shape) == 0:
            raise TypeError("Attempt to take len() of scalar dataset")
        return shape[0]

    def __iter__(self) -> Iterator[np.ndarray]:
        """ Iterate over the first axis.  TypeError if scalar.
        BEWARE: Modifications to the yielded data are *NOT* written to file.
        """
        for i in range(self.len()):
            yield self[i]

    @abstractmethod
    def __getitem__(self, args) -> np.ndarray:
        """Read a slice from the HDF5-like dataset.

        See h5py_like.shape_utils.getitem for a utility function which supports
        positive and negative integers/slices, striding, ellipses (explicit and
        implicit), reading scalars, and reading arrays with 0-length dimensions.

        Don't forget to use ``self._astype or self.dtype`` when setting the read dtype,
        to be compatible with the ``Dataset.astype`` context manager.

        The _getitem method is a helper for this.
        """

    @abstractmethod
    def __setitem__(self, args, val):
        """ Write to the HDF5-like dataset from a Numpy array.

        The _setitem method is a helper for this.
        """

    def read_direct(self, dest, source_sel=None, dest_sel=None):
        """ Read data directly from the underlying array into an existing NumPy array.
        The destination array must be C-contiguous and writable.
        Selections must be the output of numpy.s_[<args>].

        The default implementation just reads and writes as usual, and therefore has no
        speedups.
        """
        if source_sel is None:
            source_sel = Ellipsis

        if dest_sel is None:
            dest[...] = self[source_sel]
        else:
            dest[dest_sel] = self[source_sel]

    @mutation
    def write_direct(self, source, source_sel=None, dest_sel=None):
        """ Write data directly to HDF5 from a NumPy array.
        The source array must be C-contiguous.  Selections must be
        the output of numpy.s_[<args>].

        The default implementation just reads and writes as usual, and therefore has no
        speedups.
        """
        if source_sel is None:
            source_sel = Ellipsis

        if dest_sel is None:
            self[...] = source[source_sel]
        else:
            self[dest_sel] = source[source_sel]

    def __array__(self, dtype=None):
        """ Create a Numpy array containing the whole dataset.  DON'T THINK
        THIS MEANS DATASETS ARE INTERCHANGEABLE WITH ARRAYS.  For one thing,
        you have to read the whole dataset every time this method is called.
        """
        arr = self[...]
        if dtype is not None:
            arr = arr.astype(dtype)
        return arr

    def __str__(self):
        return (
            f"<{classname(self)}(name='{self.name}', "
            f"shape={self.shape}, dtype={self.dtype}, file={self.file})>"
        )

    def __eq__(self, other):
        return all(
            (
                isinstance(other, type(self)),
                self.name == other.name,
                self.parent == other.parent,
            )
        )
